"""
PDB Parser (pdb_eda.pdbParser)
-------------------------------------------------------

This module provides methods to read and parse the PDB format files and returns PDB objects.
Format details of PDB can be found in ftp://ftp.wwpdb.org/pub/pdb/doc/format_descriptions/Format_v33_Letter.pdf.
"""
import re
import numpy as np


def readPDBfile(file):
    """Creates :class:`pdb_eda.pdbParser.PDBentry` object from file name.

    :param file: The name of a PDB formated file or a file handle.
    :type file: :py:class:`str` or :class:`io.IOBase`
    """
    if isinstance(file,str):
        with open(file, "r") as fileHandle:
            return parse(fileHandle)
    else:
        return parse(file)

def parse(handle, mode='lite'):
    """Creates :class:`pdb_eda.pdbParser.PDBentry` object from file handle object.

    :param :class:`io.IOBase` handle: The file handle of a PDB formatted file.
    :param :py:class:`str` mode: Whether of not to parse all the atoms, default as 'lite' (not parse).
    :return: pdbEntry
    :rtype: :class:`pdb_eda.pdbParser.PDBentry`
    """
    atoms = []
    rotationMats = []
    modelCount = 0
    pdbid = date = method = resolution = rValue = rFree = program = spaceGroup = 0
    for record in handle.readlines():
        if mode == 'lite' and record.startswith('ATOM'):
            break
        elif record.startswith('HEADER'):
            date = record[57: 57+2].strip()
            pdbid = record[62: 62+4].strip()
        elif record.startswith('EXPDTA'):
            method = record[6: 6+30]
            method = method.strip().replace(' ', '_')
        elif record.startswith('REMARK   2 RESOLUTION'):
            match = re.search('RESOLUTION.(.+)ANGSTROMS', record)
            if match:
                resolution = match.group(1).strip()
        elif record.startswith('REMARK   3   R VALUE'):
            match = re.search('^REMARK   3   R VALUE            \(WORKING SET\) : (.+)$', record)
            if match:
                rValue = match.group(1).strip()
        elif record.startswith('REMARK   3   FREE R VALUE'):
            match = re.search('^REMARK   3   FREE R VALUE                     : (.+)$', record)
            if match:
                rFree = match.group(1).strip()
        elif record.startswith('REMARK   3   PROGRAM'):
            match = re.search('^REMARK   3   PROGRAM     : (.+)$', record)
            if match:
                program = match.group(1).strip().replace(' ', '_')
        elif record.startswith('MODEL'):
            modelCount += 1
            if modelCount > 1: break
        elif record.startswith('REMARK 290 SYMMETRY OPERATORS FOR SPACE GROUP:'):
            match = re.search('^REMARK 290 SYMMETRY OPERATORS FOR SPACE GROUP: (.+)$', record)
            if match:
                spaceGroup = match.group(1).strip().replace(' ', '_')
        elif record.startswith('REMARK 290   SMTRY'):
            match = re.search('^REMARK 290   SMTRY(.+)$', record)
            if match:
                items = match.group(1).split()
                if len(rotationMats) < int(items[1]):
                    rotationMats.append(np.zeros((3, 4)))
                rotationMats[int(items[1])-1][int(items[0])-1] = [float(i) for i in items[2:6]]
        elif record.startswith('ATOM') or record.startswith('HETATM'):
            keyValues = {'record': record,
                         'recordType': record[0: 0+6],
                         'serial': record[6: 6+5],
                         'atomName': record[12: 12+4],
                         'alternateLocation': record[16: 16+1],
                         'residueName': record[17: 17+3],
                         'chainID': record[21: 21+1],
                         'residueNumber': record[22: 22+4],
                         'x': record[30: 30+8],
                         'y': record[38: 38+8],
                         'z': record[46: 46+8],
                         'occupancy': record[54: 54+6],
                         'bFactor': record[60: 60+6],
                         'element': record[76: 76+2]}

            keyValues = {key: value.strip() for (key, value) in keyValues.items()}
            atoms.append(Atom(keyValues))

    header = PDBheader(pdbid, date, method, resolution, rValue, rFree, program, spaceGroup, rotationMats)
    return PDBentry(header, atoms)


class PDBentry:
    """:class:`pdb_eda.pdbParser.PDBentry` class that stores the :class:`pdb_eda.pdbParser.PDBheader` and/or :class:`pdb_eda.pdbParser.Atom` class."""

    def __init__(self, header, atoms):
        """:class:`pdb_eda.pdbParser.PDBentry` initializer.

        :param :class:`pdb_eda.pdbParser.PDBheader` header:
        :param :py:class:`list` atoms: list of  :class:`pdb_eda.pdbParser.Atom` objects
        """
        self.header = header
        self.atoms = atoms


class PDBheader:
    """:class:`pdb_eda.pdbParser.PDBheader` that stores information about PDB header."""

    def __init__(self, PDBid, date, method, resolution, rValue, rFree, program, spaceGroup, rotationMats):
        """:class:`pdb_eda.pdbParser.PDBheader` initializer.

        :param :py:class:`str` pdbid: PDB id.:param str pdbid: PDB id.
        :param :py:class:`str` date: PDB structure publish date.
        :param :py:class:`str` method: Experiment method, i.e. X-ray, NMR, etc.
        :param :py:class:`float` resolution: Structure resolution if applicable.
        :param :py:class:`float` rValue: Structure's R value.
        :param :py:class:`float` rFree: Structure's R free value.
        :param :py:class:`str` program: Software for acquiring the structure.
        :param :py:class:`str` spaceGroup: Structure's space group if applicable.
        :param :py:class:`list` rotationMats: Structure's rotation matrix and translation matrix if applicable.
        """
        self.pdbid = PDBid
        self.date = date
        self.method = method
        self.resolution = resolution
        self.rValue = rValue
        self.rFree = rFree
        self.program = program
        self.spaceGroup = spaceGroup
        self.rotationMats = rotationMats


class Atom:
    """:class:`pdb_eda.pdbParser.Atom` that stores information about PDB atoms."""

    def __init__(self, keyValues):
        """:class:`pdb_eda.pdbParser.Atom` initializer.

        :param :py:class:`dict` keyValues: key-value pairs for atom information.
        """
        self.record = keyValues['record']
        self.recordType = keyValues['recordType']
        self.serial = keyValues['serial']
        self.atomName = keyValues['atomName']
        self.alternateLocation = keyValues['alternateLocation']
        self.residueName = keyValues['residueName']
        self.chainID = keyValues['chainID']
        self.residueNumber = keyValues['residueNumber']
        self.x = keyValues['x']
        self.y = keyValues['y']
        self.z = keyValues['z']
        self.occupancy = keyValues['occupancy']
        self.bFactor = keyValues['bFactor']
        self.element = keyValues['element']







# !/usr/bin/python3
"""
pdb_eda single structure analysis mode command-line interface
  Analyzes a single pdb entry.

Usage:
    pdb_eda single -h | --help
    pdb_eda single <pdbid> <out-file> map (--density | --diff-density)
    pdb_eda single <pdbid> <out-file> density (--atom | --residue | --domain) [--out-format=<format>] [--params=<params-file>] [--include-pdbid]
    pdb_eda single <pdbid> <out-file> difference (--atom | --residue | --symmetry-atom) [--type=<type>] [--radius=<radius>] [--num-sd=<num-sd>] [--out-format=<format>] [--params=<params-file>] [--include-pdbid]
    pdb_eda single <pdbid> <out-file> blob [--green] [--red] [--num-sd=<num-sd>] [--out-format=<format>] [--params=<params-file>] [--include-pdbid]
    pdb_eda single <pdbid> <out-file> blob --blue [--num-sd=<num-sd>] [--out-format=<format>] [--params=<params-file>] [--include-pdbid]
    pdb_eda single <pdbid> <out-file> statistics [--out-format=<format>] (--atom | --residue) [--include-pdbid] [--print-validation]

Options:
    -h, --help                      Show this screen.
    <pdbid>                         The PDB ID to download and analyze.
    <out-file>                      Output filename. "-" will write to standard output.
    --params=<params-file>          Overriding parameters file that includes radii, slopes, etc. [default: ]
    --include-pdbid                 Include PDB ID at the beginning of each result.
    --density                       Output the density Fo-Fc map in jsonpickle format.
    --diff-density                  Output the difference density 2Fo-Fc map in jsonpickle format.
    --atom                          Calculate results for each atom.
    --residue                       Calculate results for each residue.
    --symmetry-atom                 Calculate results for each symmetry atom.
    --domain                        Calculate results for each domain.
    --green                         Calculate green (positive) difference map blobs.
    --red                           Calculate red (negative) difference map blobs.
    --blue                          Calculate blue (positive) density map blobs. Default option if red/green not selected.  However, this option uses a LOT OF MEMORY at 1.5sd.
    --radius=<radius>               Radius (in angstroms) around atom or residue to calculate significant discrepancy. [default: 3.5]
    --num-sd=<num-sd>               Number of standard deviation units to use as a significant discrepancy cutoff. Default is 3.0 for atom, residue, green, and red.  Default is 1.5 for blue.
    --type=<type>                   Residue type or atom type to filter by.
    --out-format=<format>           Output file format, available formats: csv, json [default: json].
    --print-validation              Print comparison of median absolute values below 1 sigma for Fo and Fc maps, which should be very similar.
"""

import docopt
import sys
import json
import jsonpickle
import numpy

from . import densityAnalysis
from . import __version__

def main():
    args = docopt.docopt(__doc__, version=__version__)
    if args["--help"]:
        print(__doc__)
        exit(0)

    args["--radius"] = float(args["--radius"])

    if args["--num-sd"] == None:
        args["--num-sd"] = 3.0 if args["--green"] or args["--red"] or args["difference"] else 1.5
    args["--num-sd"] = float(args["--num-sd"])

    if args["--params"]:
        try:
            with open(args["--params"], 'r') as paramsFile:
                params = json.load(paramsFile)
            densityAnalysis.setGlobals(params)
        except:
            sys.exit(str("Error: params file \"") + args["--params"] + "\" does not exist or is not parsable.")

    analyzer = densityAnalysis.fromPDBid(args["<pdbid>"])
    if not analyzer:
        sys.exit("Error: Unable to parse or download PDB entry or associated ccp4 file.")

    jsonType = "json"
    if args["--density"]:
        jsonType = "jsonpickle"
        result = analyzer.densityObj
    elif args["--diff-density"]:
        jsonType = "jsonpickle"
        result = analyzer.diffDensityObj
    elif args["density"]:
        analyzer.aggregateCloud()
        if args["--atom"]:
            headerList = list(map(str, list(analyzer.atomCloudDescriptions.dtype.names) + ['density_electron_ratio']))
            result = [[numpyConverter(element) for element in item] + [analyzer.densityElectronRatio] for item in analyzer.atomCloudDescriptions]
        elif args["--residue"]:
            headerList = densityAnalysis.DensityAnalysis.residueCloudHeader + ['density_electron_ratio']
            result = [list(item) + [analyzer.densityElectronRatio] for item in analyzer.residueCloudDescriptions]
        elif args["--domain"]:
            headerList = densityAnalysis.DensityAnalysis.domainCloudHeader + ['density_electron_ratio']
            result = [list(item) + [analyzer.densityElectronRatio] for item in analyzer.domainCloudDescriptions]
    elif args["difference"]:
        if args["--atom"]:
            headerList = densityAnalysis.DensityAnalysis.atomRegionDiscrepancyHeader
            result = analyzer.calculateAtomRegionDiscrepancies(args["--radius"], args["--num-sd"], args["--type"])
        elif args["--residue"]:
            headerList = densityAnalysis.DensityAnalysis.residueRegionDiscrepancyHeader
            result = analyzer.calculateResidueRegionDiscrepancies(args["--radius"], args["--num-sd"], args["--type"])
        elif args["--symmetry-atom"]:
            headerList = densityAnalysis.DensityAnalysis.symmetryAtomRegionDiscrepancyHeader
            result = analyzer.calculateSymmetryAtomRegionDiscrepancies(args["--radius"], args["--num-sd"], args["--type"])
            for atomInfo in result:
                atomInfo[4] = [val for val in atomInfo[4]]
                atomInfo[5] = [float(val) for val in atomInfo[5]]
    elif args['blob']:
        headerList = densityAnalysis.DensityAnalysis.blobStatisticsHeader
        result = []
        if args["--green"]:
            greenBlobList = analyzer.diffDensityObj.createFullBlobList(analyzer.diffDensityObj.meanDensity + args["--num-sd"] * analyzer.diffDensityObj.stdDensity)
            result.extend(analyzer.calculateAtomSpecificBlobStatistics(greenBlobList))
        if args["--red"]:
            redBlobList = analyzer.diffDensityObj.createFullBlobList(-1 * (analyzer.diffDensityObj.meanDensity + args["--num-sd"] * analyzer.diffDensityObj.stdDensity))
            result.extend(analyzer.calculateAtomSpecificBlobStatistics(redBlobList))
        if not args["--green"] and not args["--red"]: # args["--blue"] by default
            blueBlobList = analyzer.densityObj.createFullBlobList(analyzer.densityObj.meanDensity + args["--num-sd"] * analyzer.densityObj.stdDensity)
            result.extend(analyzer.calculateAtomSpecificBlobStatistics(blueBlobList))
        for blobInfo in result:
            blobInfo[9]  = [ val for val in blobInfo[9] ]
            blobInfo[10] = [ float(val) for val in blobInfo[10] ]
            blobInfo[11] = [ float(val) for val in blobInfo[11] ]
    elif args["statistics"]:
        if args["--print-validation"]:
            (medianAbsFo,medianAbsFc) = analyzer.medianAbsFoFc()
            print("Median abs Fo(<1sd):", medianAbsFo, "Median abs Fc(<1sd):", medianAbsFc,"Relative Difference:",(medianAbsFo - medianAbsFc)/max(medianAbsFo,medianAbsFc))

        if args["--residue"]:
            headerList = analyzer.residueMetricsHeaderList
            result = analyzer.residueMetrics()
        elif args["--atom"]:
            headerList = analyzer.atomMetricsHeaderList
            result = analyzer.atomMetrics()
            for atomInfo in result:
                atomInfo[4] = [x for x in atomInfo[4]]
                atomInfo[5] = [float(x) for x in atomInfo[5]]

    if args["--include-pdbid"]:
        headerList = ["pdbid"] + headerList
        result = [ [analyzer.pdbid] + row for row in result ]

    with open(args["<out-file>"], 'w') if args["<out-file>"] != "-" else sys.stdout as outFile:
        if args["--out-format"] == 'csv':
            csvResults = [','.join(map(str, row)) for row in [headerList] + result]
            print(*csvResults, sep='\n', file=outFile)
        elif jsonType == "jsonpickle":
            jsonText = jsonpickle.encode(result)
            outFile.write(jsonText)
        else:
            jsonResults = [ dict(zip(headerList, row)) for row in result ]
            print(json.dumps(jsonResults, indent=2, sort_keys=True), file=outFile)

def numpyConverter(obj):
    """Converts numpy objects to standard Python types, otherwise just return the object.

    :param obj:
    :return: object
    """
    if isinstance(obj, numpy.integer):
        return int(obj)
    elif isinstance(obj, numpy.floating):
        return float(obj)
    elif isinstance(obj, numpy.ndarray):
        return [numpyConverter(item) for item in obj]
    else:
        return obj



"""
File Utilities (pdb_eda.fileUtils)
----------------------------------

Common file utility functions used across pdb_eda's CLI.
"""

import os
import tempfile
import json

def createTempJSONFile(data, filenamePrefix):
    """Creates a temporary JSON file and returns its filename.

    :param data:  data to save into the JSON file.
    :type data: :py:class:`dict` or :py:class:`list`
    :param :py:class:`str` filenamePrefix: temporary filename prefix.
    :return: filename
    :rtype: :py:class:`str`
    """
    dirname = os.getcwd()
    filename = 0
    with tempfile.NamedTemporaryFile(mode='w', buffering=1, dir=dirname, prefix=filenamePrefix, delete=False) as tempFile:
        json.dump(data,tempFile)
        filename = tempFile.name
    return filename

